#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# PyODConverter (Python OpenDocument Converter) v1.1 - 2009-11-14
#
# This script converts a document from one office format to another by
# connecting to an OpenOffice.org instance via Python-UNO bridge.
#
# Copyright (C) 2008-2009 Mirko Nasato <mirko@artofsolving.com>
# Copyright (C) 2010 Per Øyvind Karlsen <peroyvind@mandriva.org>
# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl-2.1.html
# - or any later version.
#
DEFAULT_OPENOFFICE_HOST = "localhost"
DEFAULT_OPENOFFICE_PORT = 8100

import sys
try:
    import magic
except ImportError, e:
    magic = None
import uno, unohelper
from os.path import abspath, isfile, splitext, getsize
from time import sleep
from StringIO import StringIO
from com.sun.star.beans import PropertyValue
from com.sun.star.task import ErrorCodeIOException
from com.sun.star.connection import NoConnectException
from com.sun.star.lang import DisposedException
from com.sun.star.io import XInputStream, XOutputStream, XSeekable

FAMILY_TEXT = "Text"
FAMILY_WEB = "Web"
FAMILY_SPREADSHEET = "Spreadsheet"
FAMILY_PRESENTATION = "Presentation"
FAMILY_DRAWING = "Drawing"

#---------------------#
# Configuration Start #
#---------------------#

# see http://wiki.services.openoffice.org/wiki/Framework/Article/Filter

# most formats are auto-detected; only those requiring options are defined here
IMPORT_FILTER_MAP = {
    "txt": {
        "FilterName": "Text (encoded)",
        "FilterOptions": "utf8"
    },
    "text/plain": "txt",
    "csv": {
        "FilterName": "Text - txt - csv (StarCalc)",
        "FilterOptions": "44,34,0"
    },
    "pdf": {
        #FAMILY_TEXT: { "FilterName": "writer_pdf_import" },
        #FAMILY_DRAWING: { "FilterName": "draw_pdf_import" }
        "FilterName": "draw_pdf_import" # Only drawing seems supported for now..?
    },
    "application/pdf": "pdf"
}

EXPORT_FILTER_MAP = {
    "pdf": {
        FAMILY_TEXT: { "FilterName": "writer_pdf_Export" },
        FAMILY_WEB: { "FilterName": "writer_web_pdf_Export" },
        FAMILY_SPREADSHEET: { "FilterName": "calc_pdf_Export" },
        FAMILY_PRESENTATION: { "FilterName": "impress_pdf_Export" },
        FAMILY_DRAWING: { "FilterName": "draw_pdf_Export" }
    },
    "html": {
        FAMILY_TEXT: { "FilterName": "HTML (StarWriter)" },
        FAMILY_SPREADSHEET: { "FilterName": "HTML (StarCalc)" },
        FAMILY_PRESENTATION: { "FilterName": "impress_html_Export" }
    },
    "odt": {
        FAMILY_TEXT: { "FilterName": "writer8" },
        FAMILY_WEB: { "FilterName": "writerweb8_writer" }
    },
    "doc": {
        FAMILY_TEXT: { "FilterName": "MS Word 97" }
    },
    "rtf": {
        FAMILY_TEXT: { "FilterName": "Rich Text Format" }
    },
    "txt": {
        FAMILY_TEXT: {
            "FilterName": "Text",
            "FilterOptions": "utf8"
        }
    },
    "ods": {
        FAMILY_SPREADSHEET: { "FilterName": "calc8" }
    },
    "xls": {
        FAMILY_SPREADSHEET: { "FilterName": "MS Excel 97" }
    },
    "csv": {
        FAMILY_SPREADSHEET: {
            "FilterName": "Text - txt - csv (StarCalc)",
            "FilterOptions": "44,34,0"
        }
    },
    "odp": {
        FAMILY_PRESENTATION: { "FilterName": "impress8" }
    },
    "ppt": {
        FAMILY_PRESENTATION: { "FilterName": "MS PowerPoint 97" }
    },
    "swf": {
        FAMILY_DRAWING: { "FilterName": "draw_flash_Export" },
        FAMILY_PRESENTATION: { "FilterName": "impress_flash_Export" }
    }
}

PAGE_STYLE_OVERRIDE_PROPERTIES = {
    FAMILY_SPREADSHEET: {
        #--- Scale options: uncomment 1 of the 3 ---
        # a) 'Reduce / enlarge printout': 'Scaling factor'
        "PageScale": 100,
        # b) 'Fit print range(s) to width / height': 'Width in pages' and 'Height in pages'
        #"ScaleToPagesX": 1, "ScaleToPagesY": 1000,
        # c) 'Fit print range(s) on number of pages': 'Fit print range(s) on number of pages'
        #"ScaleToPages": 1,
        "PrintGrid": False
    }
}

#-------------------#
# Configuration End #
#-------------------#

def connect(host, port, timeout):
    localContext = uno.getComponentContext()
    resolver     = localContext.ServiceManager.createInstanceWithContext("com.sun.star.bridge.UnoUrlResolver", localContext)
    context      = None
    desktop      = None

    for i in reversed(xrange(timeout)):
        try:
            context = resolver.resolve("uno:socket,host=%s,port=%s;urp;StarOffice.ComponentContext" % (host, port))
            desktop = context.ServiceManager.createInstanceWithContext("com.sun.star.frame.Desktop", context)
            break
        except (NoConnectException, DisposedException), e:
            # Pause and try again to connect
            if i > 0:
                sys.stderr.write(".")
                sleep(1)
            else:
                raise e

    return context, desktop


def start(host, port, timeout, oooBinary, options = None, pidFile=None):
    import os
    args = [oooBinary,
            "-headless",
            "-accept=socket,host=%s,port=%s;urp;StarOffice.ComponentContext" % (host, port)]

    if options:
        options = options.split(" ")
        args.extend(options)

    try:
        pid = os.spawnve(os.P_NOWAIT, oooBinary, args, os.environ)
        if pid <= 0:
            return 1
    except Exception, e:
        sys.stderr.write(e.message)
        return 1

    try:
        context, desktop = connect(host, port, timeout)
        if pidFile:
            pf = open(pidFile, "w")
            pf.write("%d" % pid)
            pf.close()
        return 0
    except NoConnectException, e:
        from signal import SIGTERM, SIGABRT, SIGKILL
        # Try kill it gracefully (becoming harder) if it fails connecting..
        sys.stderr.write("\nOpenOffice.org instance running (pid %d), but unable to connect at %s:%s, terminating...\n" % (pid,host,port))
        for signal in SIGTERM, SIGABRT, SIGKILL:
            os.kill(pid, signal)
            p, s = os.waitpid(pid, os.WNOHANG)
            if p or s:
                sys.stderr.write("PID %d killed with signal %d" % (p, signal))
                return 7
            sleep(1)
            sys.stderr.write(".")
        return 1

    # LSB: "program is not running"
    return 7

def stop(host, port, timeout, terminate=False): # Overkill...?
    try:
        context, desktop = connect(host, port, timeout)
        enum = desktop.Components.createEnumeration()
        while enum.hasMoreElements():
            sys.stderr.write(".")
            elem = enum.nextElement()
            try:
                elem.close(False)
            except AttributeError:
                if not elem.supportsService("com.sun.star.frame.StartModule"):
                    elem.terminate()
        desktop.deactivate()
        if terminate: # FIXME
            desktop.terminate()
            return 0
        else:
            desktop.dispose()

        context.getServiceManager().dispose()
        context.dispose()
        return 0
    except NoConnectException, e:
        pass
    return 1

def status(host, port, timeout):
    try:
        context, desktop = connect(host,port,timeout)
        if desktop and context:
            return True
    except NoConnectException:
        pass
    return False

class FileInputStream(unohelper.Base, XInputStream, XSeekable):
    """ Minimal Implementation of XInputStream """ 
    def __init__(self, filename):
        self.filename = filename
        self.size = getsize(self.filename)
        self.file = open(filename, 'rb')
        
    def readBytes(self, data, n):
        data = self.file.read(n)
        return (len(data), uno.ByteSequence(data))
    
    def readSomeBytes(self, data, n):
        return self.readBytes(data, n)

    def skipBytes(self, n):
        self.file.seek (n, 1)

    def available(self):
        return self.size - self.file.tell();

    def closeInput(self):
        self.file.close()

    def seek(self, posn):
        self.file.seek(int(posn))

    def getPosition(self):
        return long(self.file.tell())

    def getLength(self):
        return long(self.size)

class FileOutputStream(unohelper.Base, XOutputStream):
    """ Minimal Implementation of XOutputStream """
    def __init__(self, filename):
        self.filename = filename
        self.file = open(filename, 'wb')

    def writeBytes(self, bytes):
        self.file.write(bytes.value)

    def closeOutput(self):
        self.file.close()

    def flush(self):
        self.file.flush()

class DocumentConversionException(Exception):

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message

class DocumentConverter:
    
    def __init__(self, host=DEFAULT_OPENOFFICE_HOST, port=DEFAULT_OPENOFFICE_PORT, timeout=1):
        try:
            self.context, self.desktop = connect(host, port, timeout)
        except NoConnectException:
            raise DocumentConversionException, "failed to connect to OpenOffice.org at %s:%s" % (host, port)

    def convert(self, inputFile, outputFile):
        if magic:
            ms = magic.open(0x400)
            ms.load()
            encoding = ms.file(inputFile)
            ms = magic.open(0x10)
            ms.load()
            mimeType = ms.file(inputFile)
            if encoding != "utf-8":
                # actual naming of encoding used here might not exactly
                # match whatever OOo uses, but seems to mostly match good
                # enough, with just as an important factor is whether it's
                # 'utf8' or not, if not OOo seems to be able to autodetect(..?)
                IMPORT_FILTER_MAP["txt"]['FilterOptions'] = encoding
        else:
            encoding = None
            mimeType = None
        
        inputStream = FileInputStream(inputFile)

        loadProperties = { "Hidden": True, "InputStream" : inputStream, "ReadOnly" : True }
        inputExt = self._getFileExt(inputFile)                                                                              
        if IMPORT_FILTER_MAP.has_key(inputExt):
            loadProperties.update(IMPORT_FILTER_MAP[inputExt])

        document = self.desktop.loadComponentFromURL('private:stream', "_blank", 0, self._toProperties(loadProperties))
        # Since we failed, we'll try again matching by mime type..
        if not document:
            if IMPORT_FILTER_MAP.has_key(mimeType):
                inputExt = IMPORT_FILTER_MAP[mimeType]
                loadProperties.update(IMPORT_FILTER_MAP[inputExt])
                loadProperties["InputStream"] = FileInputStream(inputFile)
                document = self.desktop.loadComponentFromURL('private:stream', "_blank", 0, self._toProperties(loadProperties))

        if not document:
            raise Exception, "Error making document"
        try:
            document.refresh()
        except AttributeError:
            pass

        family = self._detectFamily(document)
        self._overridePageStyleProperties(document, family)
        
        outputExt = self._getFileExt(outputFile)
        outputStream = FileOutputStream(outputFile)
        storeProperties = self._getStoreProperties(document, outputExt)
        storeProperties.update({"OutputStream": outputStream})
        
        try:
            document.storeToURL('private:stream', self._toProperties(storeProperties))
        finally:
            document.close(True)

    def _overridePageStyleProperties(self, document, family):
        if PAGE_STYLE_OVERRIDE_PROPERTIES.has_key(family):
            properties = PAGE_STYLE_OVERRIDE_PROPERTIES[family]
            pageStyles = document.getStyleFamilies().getByName('PageStyles')
            for styleName in pageStyles.getElementNames():
                pageStyle = pageStyles.getByName(styleName)
                for name, value in properties.items():
                    pageStyle.setPropertyValue(name, value)

    def _getStoreProperties(self, document, outputExt):
        family = self._detectFamily(document)
        try:
            propertiesByFamily = EXPORT_FILTER_MAP[outputExt]
        except KeyError:
            raise DocumentConversionException, "unknown output format: '%s'" % outputExt
        try:
            return propertiesByFamily[family]
        except KeyError:
            raise DocumentConversionException, "unsupported conversion: from '%s' to '%s'" % (family, outputExt)
    
    def _detectFamily(self, document):
        if document.supportsService("com.sun.star.text.WebDocument"):
            return FAMILY_WEB
        if document.supportsService("com.sun.star.text.GenericTextDocument"):
            # must be TextDocument or GlobalDocument
            return FAMILY_TEXT
        if document.supportsService("com.sun.star.sheet.SpreadsheetDocument"):
            return FAMILY_SPREADSHEET
        if document.supportsService("com.sun.star.presentation.PresentationDocument"):
            return FAMILY_PRESENTATION
        if document.supportsService("com.sun.star.drawing.DrawingDocument"):
            return FAMILY_DRAWING
        raise DocumentConversionException, "unknown document family: %s" % document

    def _getFileExt(self, path):
        ext = splitext(path)[1]
        if ext is not None:
            return ext[1:].lower()

    def _toFileUrl(self, path):
        return uno.systemPathToFileUrl(abspath(path))

    def _toProperties(self, dict):
        props = []
        for key in dict:
            prop = PropertyValue()
            prop.Name = key
            prop.Value = dict[key]
            props.append(prop)
        return tuple(props)

def main(argv=sys.argv):
    import getopt
    from string import atoi

    usage = "USAGE: %s <input-file> <output-file> [-h <host> -p <port>]" % argv[0]
    inputFile = None
    outputFile = None
    oooHost = DEFAULT_OPENOFFICE_HOST
    oooPort = DEFAULT_OPENOFFICE_PORT
    oooTimeout = None
    oooDaemon = None
    oooStop = None
    oooOptions = None
    pidFile = None
    checkStatus = False

    try:
        options, args = getopt.getopt(argv[1:], 'Hh:p:',
                ['host=', 'port=', 'help', 'status', 'timeout=', 'daemon=',
                    'options=', 'pidfile=', 'stop='])
        for opt, value in options:
            if opt in ('-h','--host'):
                oooHost = value
            if opt in ('-p','--port'):
                oooPort = value
            if opt in ('--status'):
                checkStatus = True
                if not oooTimeout:
                    oooTimeout = 2
            if opt in ('--timeout'):
                oooTimeout = atoi(value)
            if opt in ('--daemon'):
                oooDaemon = value
                if not oooTimeout:
                    oooTimeout = 6
            if opt in ('--options'):
                oooOptions = value
            if opt in ('--pidfile'):
                pidFile = value
            if opt in ('--stop'):
                oooStop = atoi(value)
                if not oooTimeout:
                    oooTimeout = 3
            if opt in ('-H','--help'):
                print "%s\n\n%s" % (usage, "HELP:...") #TODO
                return 0
        if not oooTimeout:
            oooTimeout = 1
        if not checkStatus and oooDaemon == None and oooStop == None:
            if len(args) < 2:
                raise getopt.GetoptError("Both input and output file needs to be specified")
            elif len(args) > 2:
                raise getopt.GetoptError("Only one input and one output file may be specified")
            else:
                inputFile, outputFile = args
    except getopt.GetoptError, msg:
        sys.stderr.write("%s\n\n%s\n" % (msg, usage))
        # LSB: "invalid or excess argument(s)"
        # (http://refspecs.freestandards.org/LSB_4.0.0/LSB-Core-generic/LSB-Core-generic/iniscrptact.html)
        return 2

    if checkStatus:
        if status(oooHost, oooPort, oooTimeout):
            print "OpenOffice.org is running at %s:%s" % (oooHost, oooPort)
            return 0
        else:
            print "Failed to connect to OpenOffice.org at %s:%s" % (oooHost, oooPort)
            # LSB: "program or service status is unknown"
            return 4

    if oooDaemon:
        return start(oooHost, oooPort, oooTimeout, oooDaemon, oooOptions, pidFile)

    if oooStop != None:
        return stop(oooHost, oooPort, oooTimeout, oooStop)

    if not isfile(inputFile):
        sys.stderr.write("no such input file: %s\n" % inputFile)
        return 1

    try:
        converter = DocumentConverter(host=oooHost, port=oooPort, timeout=oooTimeout)
        converter.convert(inputFile, outputFile)
        return 0
    except DocumentConversionException, exception:
        sys.stderr.write("ERROR! %s\n" % str(exception))
        return 1
    except ErrorCodeIOException, exception:
        sys.stderr.write("ERROR! ErrorCodeIOException %d\n" % exception.ErrCode)
        return 1

if __name__ == "__main__":
    sys.exit(main())

# vim:ts=4:sw=4:et
